use std::collections::HashMap;

struct File {
    pub size: u128
}

struct Node {
    name: String,
    node_children: HashMap<String, usize>,
    file_children: HashMap<String, File>,
    pub size: u128,
    parent: Option<usize>
}

impl Node {
    pub fn new() -> Node {
        Node {
            name: "/".to_string(),
            node_children: HashMap::new(),
            file_children: HashMap::new(),
            size: 0,
            parent: None
        }
    }

    pub fn new_with_parent(name: String, parent: usize) -> Node {
        Node {
            name: name,
            node_children: HashMap::new(),
            file_children: HashMap::new(),
            size: 0,
            parent: Some(parent)
        }
    }

    pub fn set_size(&mut self, size: u128) {
        self.size = size
    }

    pub fn get_nodes(&self) -> Vec<usize> {
        return self.node_children.iter().map(|node| *(node.1)).collect()
    }

    pub fn get_total_filesize(&self) -> u128 {
        self.file_children.iter().fold(0, |acc, (_, f)| acc + f.size)
    }
}

struct FileSystem {
    nodes: Vec<Node>,
    curr_node: usize
}

impl FileSystem {
    pub fn new() -> FileSystem {
        let root = Node::new();
        let mut nodes = Vec::new();
        nodes.push(root);
        FileSystem { nodes: nodes, curr_node: 0}
    }

    pub fn cd(&mut self, dirname: &String) {
        // println!("CDing from {} to {}", self.nodes[self.curr_node].name, dirname);
        let our_node = &self.nodes[self.curr_node];
        let target = our_node.node_children.get(dirname).unwrap();
        self.curr_node = *target;
        // println!("Now in dir {}", self.nodes[self.curr_node].name);
    }

    pub fn up(&mut self) {
        self.curr_node = self.nodes[self.curr_node].parent.unwrap();
    }

    pub fn mkdir(&mut self, dirname: String) {
        // println!("New dir {} in dir {}", dirname, self.nodes[self.curr_node].name);
        let newdir = Node::new_with_parent(dirname.clone(), self.curr_node);
        self.nodes.push(newdir);
        let newdir_idx = self.nodes.len() - 1;
        self.nodes[self.curr_node].node_children.insert(dirname.clone(), newdir_idx);
    }

    pub fn touch(&mut self, filename: String, filesize: u128) {
        // println!("New file {} in dir {}", filename, self.nodes[self.curr_node].name);
        let newfile = File { size: filesize };
        self.nodes[self.curr_node].file_children.insert(filename, newfile);
    }

    pub fn compute_sizes(&mut self) {
        self.compute_size_for_node(0);
    }

    pub fn compute_size_for_node(&mut self, node_id: usize) -> u128 {
        let children_size = self.nodes[node_id].get_nodes().iter().fold(0, |acc, child_id| acc + self.compute_size_for_node(*child_id));
        let this_node_size = self.nodes[node_id].get_total_filesize() + children_size;
        self.nodes[node_id].set_size(this_node_size);
        // println!("Size for dir {} is {}", self.nodes[node_id].name, this_node_size);
        this_node_size
    }

    pub fn sizes_smaller_than(&self, s: u128) -> u128 {
        self.nodes.iter().fold(0, |acc, node| if node.size < s { acc + node.size } else {acc})
    }

    pub fn smallest_bigger_than(&self, s: u128) -> u128 {
        let mut sizes: Vec<u128> = self.nodes.iter()
            .map(|node| node.size)
            .filter(|size| *size >= s).collect();
        sizes.sort();
        *(sizes.iter().next().unwrap())
    }

    pub fn total_size(&self) -> u128 {
        self.nodes[0].size
    }
}

pub fn part_one(input: &str) -> Option<u128> {
    let mut filesystem = FileSystem::new();

    for line in input.lines().skip(1) {
        if line.starts_with("$") {
            let mut linewalker = line.split(" ").skip(1);
            let command = linewalker.next().unwrap();
            // println!("Command is {}", command);
            if command == "cd" {
                let dirname = linewalker.next().unwrap().to_string();
                if dirname == ".." {
                    filesystem.up();
                } else {
                    filesystem.cd(&dirname);
                }
            } else if command == "ls" {
                continue;
            }
        } else {
            // parsing ls output
            if line.starts_with("dir") {
                let dirname = line.split(" ").skip(1).next().unwrap();
                filesystem.mkdir(dirname.to_string());
            } else {
                let mut linewalker = line.split(" ");
                let filesize: u128 = linewalker.next().unwrap().parse().unwrap();
                let filename = linewalker.next().unwrap().to_string();
                filesystem.touch(filename, filesize);
            }
        }
    }

    filesystem.compute_sizes();
    Some(filesystem.sizes_smaller_than(100001))
    
}

pub fn part_two(input: &str) -> Option<u128> {
    let mut filesystem = FileSystem::new();

    for line in input.lines().skip(1) {
        if line.starts_with("$") {
            let mut linewalker = line.split(" ").skip(1);
            let command = linewalker.next().unwrap();
            // println!("Command is {}", command);
            if command == "cd" {
                let dirname = linewalker.next().unwrap().to_string();
                if dirname == ".." {
                    filesystem.up();
                } else {
                    filesystem.cd(&dirname);
                }
            } else if command == "ls" {
                continue;
            }
        } else {
            // parsing ls output
            if line.starts_with("dir") {
                let dirname = line.split(" ").skip(1).next().unwrap();
                filesystem.mkdir(dirname.to_string());
            } else {
                let mut linewalker = line.split(" ");
                let filesize: u128 = linewalker.next().unwrap().parse().unwrap();
                let filename = linewalker.next().unwrap().to_string();
                filesystem.touch(filename, filesize);
            }
        }
    }

    filesystem.compute_sizes();
    let current_free = 70000000 - filesystem.total_size();
    let needed = 30000000 - current_free;
    Some(filesystem.smallest_bigger_than(needed))

}

fn main() {
    let input = &advent_of_code::read_file("inputs", 7);
    advent_of_code::solve!(1, part_one, input);
    advent_of_code::solve!(2, part_two, input);
}

#[cfg(test)]
mod tests {
    use super::*;

    #[test]
    fn test_part_one() {
        let input = advent_of_code::read_file("examples", 7);
        assert_eq!(part_one(&input), Some(95437));
    }

    #[test]
    fn test_part_two() {
        let input = advent_of_code::read_file("examples", 7);
        assert_eq!(part_two(&input), Some(24933642));
    }
}
