use std::{collections::HashSet, vec};

#[derive(Clone, Hash, Eq, PartialEq)]
struct Point {
    pub x: i32,
    pub y: i32
}

struct RopeGrid {
    head: Point,
    tail_positions: HashSet<(i32, i32)>,
    tails: Vec<Point>
}

impl RopeGrid {
    pub fn new(num_tails: usize) -> RopeGrid {
        return RopeGrid { 
            head: Point{x: 0, y: 0},
            tail_positions: HashSet::from_iter([(0, 0)]),
            tails: vec![Point{x: 0, y: 0}; num_tails]
        }
    }

    fn update_tail(&mut self) {
        let mut prev = &self.head;

        for tail in self.tails.iter_mut().rev() {
            if prev.x == tail.x && i32::abs(prev.y - tail.y) == 2 {
            if prev.y > tail.y {
                tail.y += 1;
            } else {
                tail.y -= 1;
            }
            } else if prev.y == tail.y && i32::abs(prev.x - tail.x) == 2 {
                if prev.x > tail.x {
                    tail.x += 1;
                } else {
                    tail.x -= 1;
                }
            } else if i32::abs(prev.x - tail.x) + i32::abs(prev.y - tail.y) > 2 {
                let x_dir = if prev.x > tail.x { 1 } else { -1 };
                let y_dir = if prev.y > tail.y { 1 } else { -1 };
                tail.x += x_dir;
                tail.y += y_dir;
            }
            prev = tail;
        }
        self.tail_positions.insert((prev.x, prev.y));
    }

    fn do_steps(&mut self, x_sign: i32, y_sign: i32, dist: i32) {
        for _ in 0..dist {
            self.head.x += x_sign;
            self.head.y += y_sign;
            self.update_tail();
        }
    }

    pub fn move_head(&mut self, moov: Move) {
        match moov {
            Move::Up(distance) => { self.do_steps(0, 1, distance)},
            Move::Down(distance) => { self.do_steps(0, -1, distance)},
            Move::Left(distance) => { self.do_steps(-1, 0, distance)},
            Move::Right(distance) => { self.do_steps(1, 0, distance)},
        }
    }

    pub fn num_tail_positions(&self) -> usize {
        self.tail_positions.len()
    }
}

enum Move {
    Right(i32),
    Left(i32),
    Up(i32),
    Down(i32)
}

impl Move {
    pub fn from_stirng(s: &str) -> Move {
        let (direction, steps) = s.split_once(" ").unwrap();
        match direction {
            "U" => { Move::Up(steps.parse().unwrap())},
            "D" => { Move::Down(steps.parse().unwrap())},
            "L" => { Move::Left(steps.parse().unwrap())},
            "R" => { Move::Right(steps.parse().unwrap())},
            _ => panic!("Unknown direction {}", direction)
        }
    }
}

pub fn part_one(input: &str) -> Option<usize> {
    let moves: Vec<Move> = input.lines().map(|line| Move::from_stirng(line)).collect();
    let mut ropegrid = RopeGrid::new(1);
    for moov in moves {
        ropegrid.move_head(moov)
    }
    Some(ropegrid.num_tail_positions())
}

pub fn part_two(input: &str) -> Option<usize> {
    let moves: Vec<Move> = input.lines().map(|line| Move::from_stirng(line)).collect();
    let mut ropegrid = RopeGrid::new(9);
    for moov in moves {
        ropegrid.move_head(moov)
    }
    Some(ropegrid.num_tail_positions())
}

fn main() {
    let input = &advent_of_code::read_file("inputs", 9);
    advent_of_code::solve!(1, part_one, input);
    advent_of_code::solve!(2, part_two, input);
}

#[cfg(test)]
mod tests {
    use super::*;

    #[test]
    fn test_part_one() {
        let input = advent_of_code::read_file("examples", 9);
        assert_eq!(part_one(&input), Some(13));
    }

    #[test]
    fn test_part_two() {
        let input = advent_of_code::read_file("examples", 9);
        assert_eq!(part_two(&input), None);
    }
}
